/* Version information  */
#define VERSION         3
#define SUBVERSION      0

/* Compilation switches */
#define USE_SHAREDMEMORY    1
#undef  USE_SHAREDMEMORY

/* maximum line length */
#define MAXLINE         1024

/* Define the number of lines which are processed simultaneously.       */
#define MAXFILLSIZE     512

/* Maximum number of child processes.                                                                   */
#define MAXCHILDREN     32

/* maximum length of an IP number as a string */
#define IPSTRLEN 16

/* Define the number of sets in the cache:  must be a power of 2        */
#define SETS    256

/* Define the number of ways in the cache:  must be a power of 2        */
#define WAY     256

/* Define the number of buckets in the direct-mapped cache              */
#define BUCKETS (SETS * WAY)

/* Macro to delete a carriage-return/linefeed                           */
#define DELETECRLF(str)         str[strcspn(str,"\r\n")] = '\0'

/* Define number of days until mandatory cache eviction                 */
#define EVICTWHEN       7

/* Date conversion factors                                              */
#define PERYEAR         32140800
#define PERMONTH        2678400
#define PERDAY          86400
#define PERHOUR         3600
#define PERMINUTE       60
#define PERSECOND       1


/*
 * struct nsrec - record of nameservice for cache linked list
 * 
 * ipnum - IP number hostname - hostname noname - nonzero if IP number has no
 * hostname, i.e. hostname=IP number
 */

struct dnsrec 
{
    int                 noname;
    unsigned long       ipnum;
    unsigned long       hits;
    unsigned long       evicts;
    unsigned long       timestamp;
    char                domain[80];
};

struct stat_s 
{
    unsigned            UseInFile       : 1;
    unsigned            UseOutFile      : 1;
    unsigned            UseErrFile      : 1;
    unsigned            UseCacheFile    : 1;
    unsigned            PrintStats      : 1;
    unsigned            Debug           : 1;
};


struct ipc_s
{
    char            linein[MAXLINE];
    char            domain[MAXLINE];
    unsigned char   ipnum[4];
    unsigned long   iplong;
    unsigned long   sortindex;
    unsigned long   cacheindex;
    char            *dnsname;
    char            *lineout;
    int             msgin;
    pid_t           pid;
};


struct  stats_s
{
    unsigned long   entries;        /* Total number of entries in log file:     Done    */
                                    /* Pre-lookup statistics                            */
    unsigned long   pre_resolved;   /* Number of pre-resolved IP addresses:     Done    */
    unsigned long   fake_ip_addr;   /* Number of entries that might be IP:      Done    */
                                    /* DNS cache statistics                             */
    unsigned long   hits;           /* Number of hits in DNS cache:             Done    */
    unsigned long   psuedo_hits;    
    unsigned long   evictions;      /* Number of DNS cache evictions:           Done    */
                                    /* DNS lookup statistics                            */
    unsigned long   lookups;        /* Number of actual DNS lookups:            Done    */
    unsigned long   resolved;       /* Number of successful DNS resolutions:    Done    */
    unsigned long   noname;         /* Number of unsuccessful DNS resolutions:  Done    */
};


struct  send_s
{
    int             n;
    unsigned char   ipnum[4];
};


struct  recv_s
{
    int     n;
    char    *name;
};


enum    MSG_ID
{
    MSG_IDLE,
    MSG_READY,
    MSG_TERMINATECHILD,
    MSG_CHILDTERMINATED,
    MSG_RESOLVEIP,
    MSG_IPRESOLVED,
    LU_RESOLVED,
    LU_LOOKUP,
    LU_CACHED,
    LU_EVICTION,
    LU_IDLE
};


extern  int                 sp[];
extern  int                 signum;
extern  int                 childindx;
extern  struct  stat_s      Status;
extern  struct  ipc_s       *children;
extern  struct  dnsrec      *dnscache;
extern  struct  shmid_ds    shmdat;
extern  struct  send_s      send_packet[];
extern  struct  recv_s      recv_packet;
extern  struct  stats_s     statistics;


extern  void            sig_handler(int);
extern  void            logchild(int);
extern  void            ParseCacheLine(char *);
extern  void            stats(FILE *);
extern  int             getblock(FILE *);
extern  int             getline(FILE *, char *, int);
extern  unsigned int    CacheLookup(struct ipc_s *);
extern  unsigned int    DNSLookup(struct ipc_s *);
extern  void            ExtractIPNUM(int);
extern  void            SortEntriesByIPNUM(int);
extern  void            FindCachedEntries(int);
extern  void            PrintStats(FILE *);
extern  void            Debug(char *);

