#include <netdb.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <unistd.h>

#include <sys/socket.h>
#include "log.h"


void    ExtractIPNUM(int NumEntries)
{
    int     i, k, l, n;
    int     ip_valid;
    char    c;


    for(i = 0; i < NumEntries; i ++)
    {
        /* Perform a down-and-dirty check for a pre-resolved IP     *
         * address.  If a non-digit appears as the first chraracter,*
         * the entry is considered to already have been resolved.   */
        children[i].iplong = 0;
        c = children[i].linein[0];
        if((c < '0') || (c > '9'))
        {
            children[i].dnsname     = NULL;
            children[i].lineout     = children[i].linein;
            children[i].cacheindex  = 0;
            children[i].msgin       = LU_RESOLVED;
            statistics.pre_resolved ++;
            continue;
        }

        /* Parse the IP address.  The IP address is value checked   *
         * against many bogus conditions.  All octets must be three *
         * characters or less.  Non-numeric characters are flagged  *
         * and rejected.  More than four octets appearing are       *
         * flagged and rejected.                                    */
        ip_valid = 1;
        for (k = 0, l = 0, n = 0; (k < 4) && ip_valid; k++, l = ++n) 
        {
            c = children[i].linein[n];
            while((c >= '0') && (c <= '9'))
            {
                n ++;
                c = children[i].linein[n];
            }
            if(((c != '.') && (c != ' ')) || ((c == ' ') && (k != 3)) || ((n-l) > 3))
            {
                ip_valid = 0;
                continue;
            }
            children[i].linein[n] = '\0';
            children[i].ipnum[k] = atoi(&children[i].linein[l]);
            children[i].linein[n] = c;
        }

        /* If the IP address is not valid, we should consider it as *
         * having already been resolved.                            */
        if(!ip_valid)
        {
            children[i].dnsname     = NULL;
            children[i].lineout     = children[i].linein;
            children[i].cacheindex  = 0;
            children[i].msgin       = LU_RESOLVED;
            statistics.fake_ip_addr ++;
            continue;
        }

        /* Regardless of whether a reverse DNS lookup will be       *
         * sucessful, the output will now appear as two separate    *
         * entries.  The first entry is the domain name.  If the    *
         * reverse DNS lookup fails, then the domain name will      *
         * point to a text representation of the IP address.  If    *
         * reverse DNS lookup is successful, then the correct       *
         * domain name is pointed to.                               */
        children[i].lineout = &children[i].linein[n];
        children[i].iplong = (children[i].ipnum[0] << 24) +
                             (children[i].ipnum[1] << 16) +
                             (children[i].ipnum[2] << 8) +
                              children[i].ipnum[3];
        children[i].cacheindex = children[i].iplong & (BUCKETS - 1);
    }

}


void FindCachedEntries(int NumEntries)
{
    char    *dom;
    unsigned long   sortindex = 0, ip = 0;
    int     i;

    dom = children[sortindex].dnsname;
    for(i = 0; i < NumEntries; i ++)
    {
        sortindex = children[i].sortindex;
        if(children[sortindex].iplong != ip)
        {
            CacheLookup(&children[sortindex]);
            ip  = children[sortindex].iplong;
            dom = children[sortindex].domain;
        }
        children[sortindex].dnsname = dom;
    }
}


void SortEntriesByIPNUM(int NumEntries) {
    int i, j, k, l, m;
    unsigned long li;
            
    unsigned long NoBottom = 0xFFFFFFFF;
    unsigned long NoTop    = 0;
    
    for(i = 0; i < NumEntries; i ++) 
    {
        li = (unsigned long) children[children[i].sortindex].iplong;
        if(li < NoBottom)
        {
            NoBottom = li;
            k = children[i].sortindex;
            for(j = i; j != 0; j --) {
                l = children[j-1].sortindex;
                children[j].sortindex = l;
            }
            children[0].sortindex = k;
            continue;
        }

        if(li >= NoTop) 
        {
            NoTop = li;
            continue;
        }

        for(j = 0; j < i; j++) {
            if(((unsigned long )children[children[j].sortindex].iplong) > li) 
            {
                m = children[i].sortindex;
                for(k = i; k > j; k --) {
                    l = children[k-1].sortindex;
                    children[k].sortindex = l;
                }
                children[j].sortindex = m;
                j = i;
            }
        }   
    }
}



/* DNSLookup - Performs the reverse DNS lookup on an IP address.        */
unsigned int
DNSLookup(struct ipc_s *crec)
{
    struct hostent *hostdata;
    unsigned int    indx, rtn;

    /* Calculate cache table index      */
    indx = crec->iplong & (BUCKETS - 1);
    rtn  = indx;

    /* Entry not found.  Put entry in cache.  Mark old entry as evicted */
    dnscache[indx].ipnum = crec->iplong;
    dnscache[indx].noname      = 0;
    dnscache[indx].hits        = 0;
//    dnscache[indx].timestamp   = ts;
    dnscache[indx].evicts      ++;

    /* Look up domain name from IP address      */
    hostdata = gethostbyaddr((const char *) crec->ipnum, 4, AF_INET);
    if(hostdata != NULL) 
    {
        /* Found domain name, put into cache    */
        strcpy(dnscache[indx].domain, hostdata->h_name);
    } 
    else 
    {
        /* Domain name not found.  Mark as not found.   */
        dnscache[indx].noname = 1;
        sprintf(dnscache[indx].domain, "%d.%d.%d.%d",
                crec->ipnum[0], crec->ipnum[1], crec->ipnum[2], crec->ipnum[3]);
        rtn = 0;
    }
    strcpy(crec->domain, dnscache[indx].domain);
    return(rtn);
}


/* CacheLookup - Checks to see if a specific IP address is contained    *
 * in the local DNS cache.  If it is, then the index is return, and     *
 * the DNS cache pointer record is updated to the DNS name entry.       */
unsigned int
CacheLookup(struct ipc_s *crec)
{
    unsigned long   indx;

    /* Calculate cache table index      */
    indx = crec->cacheindex;

    if(dnscache[indx].ipnum == crec->iplong)
    {
        strcpy(crec->domain, dnscache[indx].domain);
        crec->msgin     = LU_CACHED;
        statistics.hits ++;
        dnscache[indx].hits ++;
//        dnscache[indx].timestamp = ts;
        return(indx);
    }
    if(dnscache[indx].ipnum != 0)
    {
        crec->msgin = LU_EVICTION;
        statistics.evictions ++;
        return(-indx);
    }
    crec->msgin = LU_LOOKUP;
    statistics.lookups ++;
    return(0);
}



/* ParseCacheLine - receives a cache line, parses it, and populates the
 * DNS cache.
 */
void ParseCacheLine(char *line)
{
    unsigned char   ipnum[4];
    char        *lineptr;
    int         j, indx;
    unsigned long li, iplong;

    /* Format of log file entries:                                      *
     * Entry :  Hits  Evicts  TimeStamp        IP number * hostname     *
     * Finds the line number entry                                      */
    DELETECRLF(line);
    j = strspn(line, " ");

    /* Finds the number of "hits"                       */
    lineptr = index(&line[j], ' ');
    j = lineptr - line;
    j += strspn(&line[j], " ");

    /* Finds the number of evictions                    */
    lineptr = index(&line[j], ' ');
    j = lineptr - line;
    j += strspn(&line[j], " ");

    /* Find the timestamp                               */
    lineptr = index(&line[j], ' ');
    j = lineptr - line;
    j += strspn(&line[j], " ");
    li = strtoul(&line[j], (char **)NULL, 10);

    /* Find the beginning of the IP address             */
    lineptr = index(&line[j], ' ');
    j = lineptr - line;
    j += strspn(&line[j], " ");

    /* Find first octet of the IP address               */
    ipnum[0] = atoi(&line[j]);
    lineptr = index(&line[j], '.');
    j = (lineptr - line) + 1;
    
    /* Find second octet of the IP address               */
    ipnum[1] = atoi(&line[j]);
    lineptr = index(&line[j], '.');
    j = (lineptr - line) + 1;
    
    /* Find third octet of the IP address               */
    ipnum[2] = atoi(&line[j]);
    lineptr = index(&line[j], '.');
    j = (lineptr - line) + 1;
    
    /* Find fourth octet of the IP address               */
    ipnum[3] = atoi(&line[j]);
    lineptr = index(&line[j], ' ');
    j = (lineptr - line) + 1;
    
    /* Create ipnum as long integer                     */
    iplong = (ipnum[0] << 24) + (ipnum[1] << 16) + 
             (ipnum[2] << 8)  + ipnum[3];

    /* Calculate cache table index      */
    indx = iplong & (BUCKETS - 1);
    dnscache[indx].timestamp = li;
    
    /* Populate DNS cache IP address entry              */
    dnscache[indx].ipnum = iplong;
    if(line[j] == '!')
    {
        dnscache[indx].noname = 1;
    }
    
    lineptr = index(&line[j], ' ');
    j = (lineptr - line) + 1;

    /* Populate DNS cache name entry                    */
    strcpy(dnscache[indx].domain, &line[j]);
}


void PrintStats(FILE *output)
{
    fprintf(output, "\n");
    fprintf(output, "Logresolve (v3.0) statistics:\n\n");
    fprintf(output, "Already resolved in log file    :  %10ld\n", statistics.pre_resolved);
    fprintf(output, "Resolved from fake IP addresses :  %10ld\n", statistics.fake_ip_addr);
    fprintf(output, "Successful cache look-ups:         %10ld\n", statistics.hits);
    fprintf(output, "Resolved without a cache look-up:  %10ld\n", statistics.psuedo_hits);
    fprintf(output, "Cache evictions then DNS look-up:  %10ld\n", statistics.evictions);
    fprintf(output, "Successful DNS look-ups:           %10ld\n", statistics.resolved);
    fprintf(output, "Unsuccessful DNS look-ups:         %10ld\n", statistics.noname);
    fprintf(output, "===============================================\n");
    fprintf(output, "Total pre-resolved entries:        %10ld\n", statistics.pre_resolved + statistics.fake_ip_addr);
    fprintf(output, "Total cache hits:                  %10ld\n", statistics.hits + statistics.psuedo_hits);
    fprintf(output, "Total DNS look-ups:                %10ld\n", statistics.lookups + statistics.evictions);
    fprintf(output, "===============================================\n");
    fprintf(output, "Total log file entries:            %10ld\n\n", statistics.entries);
}


/*
 * prints various statistics to output
 */
void stats(FILE *output)
{
    int             i;
    char ipstring[IPSTRLEN];
    /* Print title and general statistics       */
    PrintStats(output);
    fprintf(output, "\n");

    /* Dump entire cache contents               */
    fprintf(output, "BEGIN_DNS_CACHE\n");
    fprintf(output, " Entry :  Hits  Evicts  TimeStamp        IP number * hostname\n");
    for (i = 0; i < BUCKETS; i++) 
    {
        if(dnscache[i].ipnum && (dnscache[i].evicts + 1))
        {
            sprintf(ipstring, "%d.%d.%d.%d", 
                    (unsigned int )((dnscache[i].ipnum >> 24) & 0xFF),
                    (unsigned int )((dnscache[i].ipnum >> 16) & 0xFF),
                    (unsigned int )((dnscache[i].ipnum >> 8)  & 0xFF),
                    (unsigned int )((dnscache[i].ipnum >> 0)  & 0xFF));
            fprintf(output, " %5d  %6li  %4li   %10li  %15s %s %s\n",
                    i, dnscache[i].hits, dnscache[i].evicts, dnscache[i].timestamp,
                    ipstring, dnscache[i].noname ? "!" : "-", dnscache[i].domain);
        }
    }
    fprintf(output, "END_DNS_CACHE\n");
}


int getblock(FILE *in)
{
    int i;

    for(i = 0; i < MAXFILLSIZE; i ++)
    {
        if(getline(in, children[i].linein, MAXLINE))
            break;
        children[i].sortindex   = i;
        children[i].msgin       = LU_IDLE;
        statistics.entries      ++;
    }
    return(i);
}


/*
 * gets a line from stdin
 */
int getline(FILE *in, char *s, int n)
{
    int     stcont = 1;
    while(stcont)
    {
        if(!fgets(s, n, in))
            return(1);
        DELETECRLF(s);
        if(!*s)
            continue;
        stcont = 0;
    }
    return(0);
}


void    sig_handler(int signame)
{
    int i;
    pid_t   pid = getpid();

    for(i = 0; i < MAXCHILDREN; i ++)
    {
        if(children[i].pid == pid)
            break;
    }

    switch(signame)
    {
    case SIGQUIT :
        children[i].msgin = MSG_TERMINATECHILD;
        Debug("SIGQUIT:  Terminating a child (sig_handler())...\n");
        break;
    }
}

void    Debug(char * msg)
{
    if(Status.Debug)
    {
        fprintf(stdout, "%s", msg);
        fflush(stdout);
    }
}
